import threading
from queue import Queue
from typing import Callable, cast, Optional

from pyrtition.topic.topic_message import TopicMessage
from pyrtition.topic.topic_partition import TopicPartition


class TopicPartitionConsumerThread(threading.Thread):
    number: int
    running: bool = False

    _stop_event: threading.Event
    _signal = threading.Condition
    _topic_partition: TopicPartition
    _queue: Queue
    _on_message: Callable[[TopicMessage, int, int], None] = None
    _thread_id: int
    _use_signals: bool = False

    def __init__(self, topic_partition: TopicPartition,
                 on_message: Optional[Callable[[TopicMessage, int, int], None]] = None,
                 use_signals: bool = False):
        super(TopicPartitionConsumerThread, self).__init__()
        self.setDaemon(True)
        self.setName(f"{topic_partition.topic_name}-{topic_partition.number}")
        self._use_signals = use_signals
        self._signal = threading.Condition()
        self._topic_partition = topic_partition
        self._queue = topic_partition.get_queue()
        self.number = topic_partition.number
        self.running = False
        if on_message:
            self._on_message = on_message

    def notify(self):
        if self._use_signals:
            try:
                with self._signal:
                    self._signal.notify()
            except RuntimeError as ex:
                # We have tried to notify the signal when it is being re-acquired
                pass

    def run(self) -> None:
        self.thread_id = threading.get_ident()
        self.running = True

        if self._use_signals:
            self.with_with_signals()
        else:
            self.run_without_signals()

    def run_without_signals(self):
        while self.running:
            message = self._queue.get()
            if message and self._on_message:
                try:
                    self._on_message(message, self.number, self.thread_id)
                except Exception as ex:
                    # TODO add logging
                    print(ex)
                    pass
            self._queue.task_done()

    def with_with_signals(self):
        while self.running:
            self._signal.acquire()
            try:
                while self._queue.empty():
                    self._signal.wait()
                message = self._queue.get()
                if message and self._on_message:
                    try:
                        self._on_message(message, self.number, self.thread_id)
                    except Exception as ex:
                        # TODO add logging
                        pass
            finally:
                self._signal.release()

    def stop(self):
        self.running = False




from typing import Dict, Callable

from pyrtition.consumer.topic_partition_consumer_thread import TopicPartitionConsumerThread
from pyrtition.topic.topic_coordinator import TopicCoordinator
from pyrtition.topic.topic_message import TopicMessage


class ThreadedTopicConsumer:
    _topic: TopicCoordinator
    _threads: Dict[int, TopicPartitionConsumerThread]

    on_message: Callable[[TopicMessage, int, int], None]

    def __init__(self, topic: TopicCoordinator):
        self._topic = topic
        self._threads = dict()

    def start(self):
        for partition in self._topic.partitions.values():
            thread = TopicPartitionConsumerThread(partition, self.on_message)
            self._threads[partition.number] = thread

        for thread in self._threads.values():
            thread.start()

    def stop(self):
        for thread in self._threads.values():
            thread.stop()
            thread.join()
            del thread



from datetime import datetime
from queue import Queue
from threading import RLock
from typing import Set, Any, Optional, Callable

from pyrtition.topic.topic_message import TopicMessage


class TopicPartition:
    topic_name: str
    number: int
    producer_count: int
    producers: Set[str]
    _queue: Queue
    _lock: RLock
    _consumer_thread: Any = None

    def __init__(self, name: str, number: int):
        self.topic_name = name
        self.number = number
        self.producer_count = 0
        self.producers = set()
        self._queue = Queue()
        self._lock = RLock()

    def dequeue(self) -> Optional[TopicMessage]:
        if not self._queue.empty():
            return self._queue.get()

    def is_queue_empty(self) -> bool:
        return self._queue.empty()

    def get_queue(self) -> Queue:
        return self._queue

    def has_producer(self, producer_name):
        return producer_name in self.producers

    def assign_producer(self, producer_name) -> int:
        self._lock.acquire()
        try:
            self.producers.add(producer_name)
            self.producer_count = len(self.producers)
            return self.producer_count
        finally:
            self._lock.release()

    def put_value(self, producer_name: str, data: Any):
        if not self.has_producer(producer_name):
            raise Exception(f"Producer {producer_name} is not a member of this partition")

        message = TopicMessage(producer_name=producer_name, timestamp=datetime.utcnow(), data=data)
        self._queue.put(message)

        # Notify any consumers
        if self._consumer_thread:
            self._consumer_thread.notify()

    def start_consuming(self, on_message: Callable[[TopicMessage, int, int], None] = None, use_signals: bool = False):
        # We can only consume if we have an on_message callable
        if not on_message:
            return

        from pyrtition.consumer.topic_partition_consumer_thread import TopicPartitionConsumerThread
        consumer_thread = TopicPartitionConsumerThread(self, on_message, use_signals)
        consumer_thread.start()
        self._consumer_thread = consumer_thread

    def stop_consuming(self):
        if self._consumer_thread:
            self._consumer_thread.stop()


from queue import Queue
from threading import RLock
from typing import Dict, Callable, Optional

from pyrtition.topic.topic_coordinator import TopicCoordinator
from pyrtition.topic.topic_message import TopicMessage


class Coordinator:
    topics: Dict[str, TopicCoordinator]

    _lock: RLock

    def __init__(self):
        self.topics = dict()
        self._lock = RLock()

    def get_topic(self, topic_name: str) -> Optional[TopicCoordinator]:
        if topic_name in self.topics:
            return self.topics[topic_name]
        return None

    def get_or_create_topic(self, topic_name: str, max_partition_count: int = 4) -> TopicCoordinator:
        if topic_name in self.topics:
            return self.topics[topic_name]
        new_topic: TopicCoordinator
        try:
            self._lock.acquire()
            new_topic = TopicCoordinator(topic_name, max_partition_count)
            self.topics[topic_name] = new_topic
        finally:
            self._lock.release()
        return new_topic

    def publish(self, topic_name, producer_name, value: any, create_if_not_exists: bool = False) -> bool:
        if create_if_not_exists:
            topic = self.get_or_create_topic(topic_name)
        else:
            topic = self.get_topic(topic_name)
        if topic:
            return topic.publish(producer_name, value)
        else:
            return False

    def get_topic_partition_queue(self, topic_name: str, partition: int) -> Queue:
        topic = self.get_or_create_topic(topic_name)
        return topic.get_queue(partition)

    def create_and_start_consuming(self, topic_name: str, max_partition_count: int = 4,
                                   on_message: Callable[[TopicMessage, int, int], None] = None) -> TopicCoordinator:
        topic = self.get_or_create_topic(topic_name, max_partition_count)
        if on_message:
            topic.start_consuming(on_message)
        return topic



import random
from queue import Queue
from threading import RLock
from typing import Dict, List, Any, Optional, Callable

from pyrtition.topic.topic_message import TopicMessage
from pyrtition.topic.topic_partition import TopicPartition
from pyrtition.topic.topic_partition_capacity import TopicPartitionCapacity


class TopicCoordinator:
    name: str
    max_partition_count: int

    partitions: Dict[int, TopicPartition]
    _producer_cache: Dict[str, int]
    _lock: RLock

    def __init__(self, name="default", max_partition_count: int = 4):
        self.name = name
        self.max_partition_count = max_partition_count
        self._lock = RLock()
        self.__partition()

    def __partition(self):
        self.partitions = dict()
        self.producer_cache = dict()
        for i in range(self.max_partition_count):
            self.__create_partition(i + 1)

    def __create_partition(self, number: int) -> bool:
        partition = TopicPartition(self.name, number)
        self.partitions[number] = partition
        return True

    def __assign_new_producer_to_partition(self, producer_name: str) -> int:
        self._lock.acquire()
        try:
            next_available_partition = self.__get_next_available_partition()
            if next_available_partition is not None:
                partition = self.partitions[next_available_partition]
                partition.assign_producer(producer_name)
                self.producer_cache[producer_name] = partition.number
                return partition.number
        finally:
            self._lock.release()

    def __get_next_available_partition(self):
        capacities = self.get_capacity()
        # Sort them in descending order
        capacities.sort(key=lambda c: c.producers)

        # If there is only 1 available then just return it
        if len(capacities) == 1:
            return capacities[0].number

        # Pick the next available zero capacity partition
        next_available_zero_capacity = next((c for c in capacities if c.producers == 0), None)
        if next_available_zero_capacity is not None:
            return next_available_zero_capacity.number

        # Either pick the lowest available partition or a random one
        pick_lowest = random.getrandbits(1)
        if pick_lowest:
            return capacities[0].number
        else:
            # Pick a random low capacity topic
            next_available_random = random.randint(0, 3)
            if len(capacities) <= next_available_random:
                return capacities[next_available_random].number

        # As a last resort just return the first partition
        return capacities[0].number

    def get_producer_partition(self, producer_name: str) -> int:
        if producer_name in self.producer_cache:
            return self.producer_cache[producer_name]
        raise Exception(f"Producer {producer_name} is not in topic {self.name}")

    def get_or_add_producer_partition(self, producer_name: str) -> int:
        if producer_name in self.producer_cache:
            return self.producer_cache[producer_name]
        else:
            return self.__assign_new_producer_to_partition(producer_name)

    def publish(self, producer_name: str, data: Optional[Any]) -> bool:
        assigned_to = self.get_or_add_producer_partition(producer_name)

        if data is not None and assigned_to > 0:
            self.partitions[assigned_to].put_value(producer_name, data)
            return True
        return False

    def is_queue_empty(self, partition: int) -> bool:
        if self.partitions[partition]:
            return self.partitions[partition].is_queue_empty()

    def dequeue(self, partition: int) -> Any:
        if self.partitions[partition]:
            return self.partitions[partition].dequeue()

    def get_queue(self, partition: int) -> Queue:
        if self.partitions[partition]:
            return self.partitions[partition].get_queue()

    def get_capacity(self) -> List[TopicPartitionCapacity]:
        return list([TopicPartitionCapacity(partition.number, partition.producer_count)
                     for partition in self.partitions.values()])

    def start_consuming(self, on_message: Callable[[TopicMessage, int, int], None] = None, use_signals: bool = False):
        for partition in self.partitions.values():
            partition.start_consuming(on_message, use_signals)

    def stop_consuming(self):
        for partition in self.partitions.values():
            partition.stop_consuming()



from setuptools import setup, find_packages


setup(
    name="pyrtition",
    version="1.0.2",
    packages=find_packages(exclude="tests"),
    license='MIT',
    author='Jeremy Carter',
    author_email='jeremychild@gmail.com',
    python_requires='>=3.6',
    url="https://github.com/jeremylcarter/pyrtition"
)